import boto3
import warnings
import ipdb
warnings.simplefilter(action='ignore')
import pandas as pd
import humanize

def parse_keys(x):
    try:
        return [ r['Value'] for r in x if r['Key']=='Name'][0]
    except:
        print("one of the instances does not have a name")

def get_ec2(status: str, profile: str, filter: str):
    possible_filters = dict(
        live=[{"Name":"instance-state-name", "Values":["running"] }],
        all=[]
    )

    filters = possible_filters[filter]

    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')
    live_inst = ec2.describe_instances(Filters=filters)
    df = pd.DataFrame(live_inst['Reservations'])
    try:
        df_list = [ pd.DataFrame.from_dict(x[0], orient='index') for x in df['Instances'].values ]
    except KeyError as e:
        return f'No instances running for {profile} with filter {filter}'
    full_df = pd.concat(df_list,axis=1, sort=True).T

    full_df['SG'] = [ ','.join([i['GroupName'] for i in f ]) for f in  full_df.SecurityGroups  ] 
    
    # humanize all LaunchTime timestamps
    # full_df['LaunchTime'] = [ humanize.naturaltime(i) for i in full_df.LaunchTime.values]  - pd.Timestamp.now()
    full_df['LaunchTime'] = full_df.LaunchTime.astype('int64').apply(lambda x: humanize.naturaldelta(pd.to_datetime(x)))

    # shorten CPU options
    full_df['CPUs'] = full_df.CpuOptions.apply(lambda x: x['CoreCount'])
    del full_df['CpuOptions']

    target_columns = 'CPUs InstanceId InstanceType LaunchTime SG PublicIpAddress Tags PrivateIpAddress'.split(' ')

    # check if tags present
    if 'Tags' not in full_df.columns:
        full_df.Tags = 'N/A'
    return_df = full_df[target_columns]

    return_df['name'] = return_df.Tags.apply(lambda x: parse_keys(x))

    pd.set_option('display.max_colwidth', 40)

    return return_df

def update_ec2_ssh(profile: str):
    from .sshconf import read_ssh_config
    # get ec2 and ssh config to get hosts
    running_instances = get_ec2("running", profile=profile, filter='live').T
    ssh_config_obj = read_ssh_config('/Users/jakub/.ssh/config')
    hosts = ssh_config_obj.hosts()
    changes = {}

    for index, row in running_instances.T.iterrows():
        try:
            target_host = [ t['Value'] for t in row.Tags if t['Key']=='Name'][0]
        except IndexError as e:
            raise IndexError("You probably did not set up the EC2 tags correctly (there are empties)")
        except TypeError as e:
            print(f"Raised TypeError {e}. Likely missing tags for row {row}")
        target_ip = row.PublicIpAddress
        # change and write
        if target_host in ssh_config_obj.hosts_:
            ssh_config_obj.set(target_host, HostName=target_ip)
            changes[target_host] = target_ip
            ssh_config_obj.write('/Users/jakub/.ssh/config')
        else:
            print(f'Could not find {target_host} in SSH config.')

    print(f"Changed {changes.keys()} to {changes.items()} respectively.")


def start_ec2(instance_id: str, profile: str):
    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')

    resp = ec2.start_instances(InstanceIds=[instance_id])
    print(resp)


def stop_ec2(inst_id: str, profile: str):
    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')

    resp = ec2.stop_instances(InstanceIds=[instance_id])
    print(resp)

def change_inst_type(inst_id: str, profile: str):
    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')

    resp = ec2.change


# pip install github3.py
import github3
import pandas as pd
import sys
import subprocess as sp

def init():
    git_config = sp.run(["git","config","-l"], capture_output=True).stdout
    split_config = git_config.decode().split("\n")
    git_dict = {s[0]:s[1] for s in [ s.split('=') for s in split_config[:-11] ]}

    gh_username = git_dict['user.email']
    gh_password = git_dict['user.token']

    return gh_username, gh_password

def find_forks(repo_url: str):
    gh_username, gh_password = init()

    user, repo = repo_url.split('/')[-2:]

    g = github3.login(gh_username, gh_password)
    r = g.repository(user, repo)
    total_list = list(r.forks(sort='commits', number=r.fork_count))
    data = [ (i,i.updated_at,
            i.pushed_at, i.updated_at == i.created_at) for i in total_list ]
    df = pd.DataFrame(data,columns=['name','updated_at','pushed_at','ever_changed'])

    # TODO: humanize dates/times
    # https://github.com/jmoiron/humanize
    df = df.sort_values('ever_changed')
    print(df.ever_changed.value_counts())
    return df


def find_contributors(repo_url: str):
    from .sshconf import SshConfig
    # gh_username, gh_password = init()

    user, repo = repo_url.split('/')[-2:]
    f = open('/Users/jakub/.gitconfig').read().splitlines()
    c = SshConfig(f)

    g = github3.login(gh_username, gh_password)
    r = g.repository(user, repo)


#/bin/python

import boto3
import pandas as pd
import os 

try:
    term_wdith = os.get_terminal_size().columns
except OSError:
    term_wdith = 70

# pd.set_option('display.max_colwidth',term_wdith - 40)
pd.set_option('display.max_colwidth',30)

def get_public_ip():
    from urllib.request import urlopen
    import re
    data = str(urlopen('http://checkip.dyndns.com/').read())
    # data = '<html><head><title>Current IP Check</title></head><body>Current IP Address: 65.96.168.198</body></html>\r\n'
    return re.compile(r'Address: (\d+\.\d+\.\d+\.\d+)').search(data).group(1)

def get_sg(sg_name: str, profile: str):
    import boto3
    sg_name = sg_name or 'default'
    profile = profile or 'default'

    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')
    resp = ec2.describe_security_groups()
    target_sg_str = list(filter(lambda x: x['GroupName']==sg_name, resp['SecurityGroups']))
    return target_sg_str

def check_if_aux(ip_permissions: list):
    try:
        return 'Aux' in ip_permissions['IpRanges'][0]['Description']
    except IndexError as e:
        assert ip_permissions['IpRanges']==[]
        return False
    except KeyError as e:
        return False

def update_sg_to_ip(sg_name='jakub', # ip='2.222.105.71/32',
                    ports=[22, 8888, 8501, 8889, 5432], profile='default'):
    import boto3

    ip = f"{get_public_ip()}/32"

    ec2 = boto3.session.Session(profile_name=profile, region_name='eu-west-1').client('ec2')
    profile = profile or 'default'

    target_sg_str = get_sg(sg_name, profile=profile)[0]
    ec2_res = boto3.session.Session(profile_name=profile, region_name='eu-west-1').resource('ec2')
    sg_res = ec2_res.SecurityGroup(target_sg_str['GroupId'])

    # loop using description to only remove Aux and not e.g. NFS or Home
    # old port-based solution
    # permissions_to_revoke = [ s for s in sg_res.ip_permissions if s['FromPort'] in ports ]
    permissions_to_revoke = [ s for s in sg_res.ip_permissions if check_if_aux(s) ]

    # Removes all relevant permissions from this group (originally as `jakub`)
    if sg_res.ip_permissions!=[] and permissions_to_revoke!=[]:
        print('No permissions to revoke. Skipping.')
        sg_res.revoke_ingress(IpPermissions=permissions_to_revoke)

    descriptions = {
        22: 'Jakub Aux SSH',
        80: 'Web',
        8888: 'Jakub Aux Jupyter',
        8501: 'Jakub Aux Streamlit',
        8889: 'Jakub Aux Jupyter2',
        5432: 'Jakub Aux Postgres'
    }
    

    ip_perms = [ {
        'IpProtocol': 'tcp',
        'FromPort': ports[i],
        'ToPort': ports[i],
        'IpRanges': [{'CidrIp': ip, 'Description': descriptions[ports[i]] }]}
            for i in range(len(ports)) ]

    # print(ip_perms)

    data = ec2.authorize_security_group_ingress(
            GroupId=target_sg_str['GroupId'],
            IpPermissions=ip_perms )
    print('Ingress Successfully Set %s' % data)

# lists all security groups
def list_security_groups():
    # define a boto3 client with a region eu-est
    ec2 = boto3.client('ec2', region_name='eu-west-1')
    response = ec2.describe_security_groups()
    security_group_df = pd.DataFrame(response['SecurityGroups'])

    # extract keys from Tags column
    # [ x for x in df.Tags if not np.all(pd.isna(x)) ] 

    # generate a mask for terms that have FILTER_TERMS in them
    FILTER_TERMS = ['LIW']
    mask = security_group_df.Description.apply(lambda x: any(term in x for term in FILTER_TERMS))
    security_group_df = security_group_df[~mask]

    # display all columns except FILTER_COLS
    FILTER_COLS = ['VpcId', 'Tags','OwnerId']
    security_group_df = security_group_df[security_group_df.columns.difference(FILTER_COLS)]

    # create an Ingress dataframe by filtering on IpPermissions not empty
    df_ingress = security_group_df[security_group_df.IpPermissions.notnull()]
    ip_permissions_df = pd.DataFrame.from_dict([x[0] for x in df_ingress.IpPermissions.values])
    
    df_ingress = df_ingress.reset_index().drop('index',axis=1)
    
    viewable_df = pd.concat([ip_permissions_df,df_ingress],axis=1).drop(['Ipv6Ranges','IpPermissions'],axis=1)
    # viewable_df.Description = viewable_df.Description.str[:30]

    # reorders columns 
    order = "GroupName FromPort IpProtocol ToPort UserIdGroupPairs PrefixListIds Description IpRanges IpPermissionsEgress GroupId".split()
    viewable_df = viewable_df[order]

    return viewable_df


# construct a DataFrame from df.IpPermissionsEgress dictionary
# ip_permissions_df = pd.DataFrame.from_dict([x[0] for x in df.IpPermissionsEgress.values])


import gpustat as gp # TODO: add to requirements
import time, os
import subprocess as sp
import sys
import importlib.util

def monitor():
    c = gp.new_query()
    processes_per_gpu = [ len(x['processes']) for x in  c.jsonify()['gpus']]
    return processes_per_gpu

def submit(job_text):
    '''
    Submits a job and keeps it in Q until GPUs have no processes:

    job_text:   e.g. `python train.py --lr 0.1`

    TODO: add shutdown as an option.
    TODO: add some indication of what runs where.
    '''
    tar_file = [ f for f in job_text.split(' ') if '.py' in f]
    torch = importlib.util.find_spec('torch') is not None
    assert len(tar_file)==1

    while True:
        ppgpu = monitor()
        if sum(ppgpu)>1:
            print(f'Still busy. File {tar_file[0]} is in local: {tar_file[0] in os.listdir()}. \
            Torch present {torch}.')
        else:
            print('Done, submitting a new job!')
            # my_env = os.environ.copy()
            sp.run(job_text, shell=True)
            time.sleep(120)
            sys.exit(0)
        time.sleep(1)




import pandas as pd
import swifter # ==0.289
import os, sys, io
import subprocess as sp

def execute(command: str, filt_cond: str, directory=''):
    '''
    Runs a python `command` on filenames that pass -f
    filt_cond: filter all files in in this directory and below by this expression.
    command: E.g.
    `dt py 'f"mv {x} real_A/ "' -f "real_A"`
    '''
    df = get_files_as_pd(directory)

    import ipdb; ipdb.set_trace()

    target_files = pd.Series(filter(lambda x: filt_cond in x, df.full_path.values))
    target_fcn = lambda x: os.system(eval(command))
    target_files.swifter.apply(lambda x: target_fcn(x))


def get_files_as_pd(directory):
    '''
    directory: `str` of full_path

    Gets all files from recursive directories and returns this as
    a `DataFrame`
    '''
    # TODO: find a way to print STDOUT and STDERR as a print statement
    try:
        proc = str(sp.check_output(f'find {directory}* -type f',
                                    shell=True,stderr=sp.PIPE))[2:]
    except sp.CalledProcessError as e:
        print(f'Error {e} Possibly incorrect path.')
        raise AssertionError
    df = pd.DataFrame(str(proc).split('\\n')[:-1] )
    df.columns = ['full_path']

    level_paths = df.full_path.str.split('/',expand=True)
    df = pd.concat([df, level_paths], axis=1)
    return df






def check_docs(variable, search, m):
    if type(getattr(variable, m).__doc__)!=str:
        return False
    else:
        return search in getattr(variable, m).__doc__

def s(variable, search: str):
    methods = dir(variable)
    return [ m for m in methods if search in m ]


def d(variable, search: str):
    methods = dir(variable)
    for m in methods:
        if check_docs(variable, search, m):
            doc = getattr(variable,m).__doc__
            tar = doc.index(search)

            # TODO: add excluded terms in some way
            print(f"\x1b[0m \x1B[1m {m}:")
            print(f"\x1b[0m{doc[tar-100:tar-5]}\x1b[31m{doc[tar-5:tar+5]}\x1b[0m{doc[tar+5:tar+100]}" )


def a(variable):
    methods = dir(variable)
    return { m:getattr(variable, m) for m in methods if not m.startswith('_') }


import sys
import gkeepapi
from datetime import datetime as dt
import pandas as pd
sys.path.append('/Users/jakub/remote_code')
from auth import email, gkeep_password


keep = gkeepapi.Keep()
success = keep.login(email, gkeep_password)

def add_todo(task: str, length_time: float, project: str):
    glist = keep.find(query='ToDoDB')
    target = next(keep.find(query='ToDoDB'))
    inserted = str(dt.now())
    completed = ''
    assert project in ['DA', 'SL']
    # last space for completed
    target.add(f"{project} || {task} || {inserted} || {length_time} || ", False)
    keep.sync()

def list_todos(done: bool, project: str):
    target = next(keep.find(query='ToDoDB'))
    df = pd.DataFrame([ i.text.split('||') for i in target.items ])
    df.columns = ['Project', 'Task','StartedAt','#hrs','FinishedAt']
    if eval(str(done)): df = df[df.FinishedAt.str.isspace()]
    if project: df = df[df.Project.str.strip()==project]
    print(df)

def done_todo(task_id: str):
    target = next(keep.find(query='ToDoDB'))
    task_id = int(task_id)
    target.items[task_id].checked = True
    done_time = dt.now()
    target.items[task_id].text += str(done_time)
    print(f"Marked as done: {target.items[task_id].text}")
    keep.sync()


import streamlit as st
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os
import cv2
from PIL import Image
import fire
import random


def process_one(image_type: str, image_name: str, percentage: float,
                image_root: str, adapted_root: str):
    # TODO: genericise. Currently assumes all file names are the same.
    target_image = f'{image_root}/{image_type}/{"_".join(image_name.split("_"))}'
    img = cv2.imread(target_image)[:,:,::-1]
    img = cv2.resize(img, (512,512))

    adapted_img = f"{adapted_root}/{image_name}"
    adapted_img = np.array(Image.open(adapted_img).resize((512,512)))
    # adapted_img = cv2.imread(adapted_img)
    adapted_img.resize((512,512,3))
    # adapted_img = adapted_img[:,:,::-1]

    img_adj = int( img.shape[1] * percentage )
    reverse_adj = img.shape[1] - img_adj
    result_img = np.hstack(( img[:,:img_adj,:], adapted_img[:, img_adj:,:] ))
    # result_img = np.concat([img[:percentage_adj], adapted_img[percentage_adj:])

    return result_img



def create_viz(
    images = '/efs/public_data/gta5/images',
    labels = 'labels',
    adapted = '/efs/public_data/images'):

    labels = '/'.join(images.split('/')[:-1]) + labels
    # set up ui
    label_paths = sorted([ x for x in os.listdir(labels) ])[:1000]
    image_paths = sorted([ x for x in os.listdir(images) ])[:1000]
    adapted_paths = sorted([ x for x in os.listdir(adapted) ])[2:1000]

    im_type = ['rgb', 'label']
    tar_im_type = st.sidebar.selectbox("Which type of image?", im_type)

    file_selection = st.sidebar.selectbox("Which file?", adapted_paths)

    percentage = st.sidebar.slider("percentage", 0, 100) / 100

    if st.button('Random Image'):
        file_selection = random.choice(image_paths)
    #     i = image_paths.index(file_selection)
    #     file_selection = image_paths[i+var_i]
    #     var_i += 1
        st.code(f"{adapted}/{file_selection}")
    else:
        st.write(f'Offset {var_i}')

    image_root = '/'.join(images.split('/')[:-1])
    adapted_root = '/'.join(adapted.split('/')[:-1])

    kpd_img = process_one(tar_im_type, file_selection, percentage,
                            image_root, adapted_root)
    kpd_img_1 = process_one(tar_im_type, file_selection, 0,
                            image_root, adapted_root)
    kpd_img_2 = process_one(tar_im_type, file_selection, 1,
                            image_root, adapted_root)

    # st.image(kpd_img, width=256)
    # st.image(kpd_img_1, width=256)
    # st.image(kpd_img_2, width=256)

    col1, col2, col3 = st.beta_columns(3)
    col1.header("Mix")
    col1.image(kpd_img, use_column_width=True)

    col2.header("Adapted")
    col2.image(kpd_img_1, use_column_width=True)

    col3.header("Original")
    col3.image(kpd_img_2, use_column_width=True)


if __name__ == "__main__":
    fire.Fire(create_viz)


'''
Stopped training alerting script.
'''
__author__ = 'Jakub Langr'
import subprocess
import sentry_sdk
import datetime
import os
from sentry_sdk import capture_exception, capture_message
from argparse import ArgumentParser
import sys
import subprocess as sp
import traceback

def initialize(overwrite=False):
    global config
    script_location = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(script_location)
    try:
        import perses_config as config
    except ModuleNotFoundError:
        if not os.path.exists(f"{script_location}/perses_config.py") or overwrite:
            print('Config does not exist. Please enter the commands.')
            exec_file = input('Enter the file to run: ')
            defaults = input("Please set default flags: ")
            sentry_sdn = input("Enter the Sentry SDN: ")
            with open(f"{script_location}/perses_config.py","w+") as f:
                f.write(f"exec_file = \'{exec_file}\' \n")
                f.write(f"defaults = \'{defaults}\' \n")
                f.write(f"sentry_sdn = \'{sentry_sdn}\' \n")

            import perses_config as config
        else:
            raise Exception('Missing config & cannot create one.')


    # variables to set
    now_str = str(datetime.datetime.now())[:-7].replace(' ','_')
    log_file = f'cli_output/{config.exec_file}_{now_str}.txt'

    # uses the current python env
    os.environ['BETTER_EXCEPTIONS'] = "1"

    # makes sure the file above exists
    # open(log_file, 'a').close()

    # This should be eventually grabbed from ENV or something.
    sentry_sdk.init(config.sentry_sdn)


def tracking(command: str, testing: bool, no_shutdown: bool, base_py: str):
    start_time = datetime.datetime.now()
    initialize()

    shutdown_hours = [22, 23] + list(range(0,8))
        
    # e.g. command = f'{base_py} train.py {FLAGS} 2>> {log_file}'
    # alt command = f'{base_py} train.py {FLAGS} 2>&1 | tee {log_file}'

    try:
        print(f'Starting tracking, running command: \n {command}')
        output = subprocess.run(
            command, shell=True, check=True) 
        # print(output.stderr)
    except KeyboardInterrupt as e:
        print('Interrupted by user. Not sending a message and not shutting down.')
    except Exception as ex:
        # TODO: maybe a better version of the tb? 
        traceback.print_exception(type(ex), ex, ex.__traceback__)
        run_time = datetime.datetime.now() - start_time 

        # send error message to sentry
        if run_time.total_seconds() > 60:
            print('Program exited, Output: \n{}'.format(ex))
            capture_message('Program exited, Output: \n{}'.format(ex))
        
        # possibly shut down the instance
        time_checks = run_time.total_seconds() > 60 and datetime.datetime.now().hour in shutdown_hours
        if not no_shutdown and time_checks:
            capture_message('It is also late. Shutting down the instance.')
            os.system('sudo shutdown now')

def test_sentry():
    initialize()
    inst_private_ip = os.environ.get("SSH_CONNECTION").split()[2]
    capture_message(f"A test if sentry can work from private IP: \
        {inst_private_ip}")




import re
from collections import defaultdict, namedtuple

# taken from "man ssh"
KNOWN_PARAMS = (
    "AddKeysToAgent",
    "AddressFamily",
    "BatchMode",
    "BindAddress",
    "CanonicalDomains",
    "CanonicalizeFallbackLocal",
    "CanonicalizeHostname",
    "CanonicalizeMaxDots",
    "CanonicalizePermittedCNAMEs",
    "CertificateFile",
    "ChallengeResponseAuthentication",
    "CheckHostIP",
    "Cipher",
    "Ciphers",
    "ClearAllForwardings",
    "Compression",
    "CompressionLevel",
    "ConnectionAttempts",
    "ConnectTimeout",
    "ControlMaster",
    "ControlPath",
    "ControlPersist",
    "DynamicForward",
    "EscapeChar",
    "ExitOnForwardFailure",
    "FingerprintHash",
    "ForwardAgent",
    "ForwardX11",
    "ForwardX11Timeout",
    "ForwardX11Trusted",
    "GatewayPorts",
    "GlobalKnownHostsFile",
    "GSSAPIAuthentication",
    "GSSAPIKeyExchange",
    "GSSAPIClientIdentity",
    "GSSAPIDelegateCredentials",
    "GSSAPIRenewalForcesRekey",
    "GSSAPITrustDns",
    "GSSAPIKexAlgorithms",
    "HashKnownHosts",
    "Host",
    "HostbasedAuthentication",
    "HostbasedKeyTypes",
    "HostKeyAlgorithms",
    "HostKeyAlias",
    "HostName",
    "IdentitiesOnly",
    "IdentityAgent",
    "IdentityFile",
    "Include",
    "IPQoS",
    "KbdInteractiveAuthentication",
    "KbdInteractiveDevices",
    "KexAlgorithms",
    "LocalCommand",
    "LocalForward",
    "LogLevel",
    "MACs",
    "Match",
    "NoHostAuthenticationForLocalhost",
    "NumberOfPasswordPrompts",
    "PasswordAuthentication",
    "PermitLocalCommand",
    "PKCS11Provider",
    "Port",
    "PreferredAuthentications",
    "email",
    "token",
    "Protocol",
    "ProxyCommand",
    "ProxyJump",
    "ProxyUseFdpass",
    "PubkeyAcceptedKeyTypes",
    "PubkeyAuthentication",
    "RekeyLimit",
    "RemoteForward",
    "RequestTTY",
    "RhostsRSAAuthentication",
    "RSAAuthentication",
    "SendEnv",
    "ServerAliveInterval",
    "ServerAliveCountMax",
    "StreamLocalBindMask",
    "StreamLocalBindUnlink",
    "StrictHostKeyChecking",
    "TCPKeepAlive",
    "Tunnel",
    "TunnelDevice",
    "UpdateHostKeys",
    "UsePrivilegedPort",
    "User",
    "UserKnownHostsFile",
    "VerifyHostKeyDNS",
    "VisualHostKey",
    "XAuthLocation"
)

known_params = [x.lower() for x in KNOWN_PARAMS]  # pylint: disable=invalid-name

class ConfigLine:  # pylint: disable=too-few-public-methods
    """ Holds configuration for a line in ssh config """
    def __init__(self, line, host=None, key=None, value=None):
        self.line = line
        self.host = host
        self.key = key
        self.value = value

def read_ssh_config(path):
    """
    Read ssh config file and return parsed SshConfig
    """
    with open(path, "r") as fh_:
        lines = fh_.read().splitlines()
    return SshConfig(lines)

def empty_ssh_config():
    """
    Creates a new empty ssh configuration.
    """
    return SshConfig([])

def _key_value(line):
    no_comment = line.split("#")[0]
    return [x.strip() for x in re.split(r"\s+", no_comment.strip(), 1)]

def _remap_key(key):
    """ Change key into correct casing if we know the parameter """
    if key in KNOWN_PARAMS:
        return key
    if key.lower() in known_params:
        return KNOWN_PARAMS[known_params.index(key.lower())]
    return key

class SshConfig(object):
    """
    Class for manipulating SSH configuration.
    """
    def __init__(self, lines):
        self.lines_ = []
        self.hosts_ = set()
        self.parse(lines)

    def parse(self, lines):
        """Parse lines from ssh config file"""
        cur_entry = None
        for line in lines:
            kv_ = _key_value(line)
            if len(kv_) > 1:
                key, value = kv_
                if key.lower() == "host":
                    cur_entry = value
                    self.hosts_.add(value)
                self.lines_.append(ConfigLine(line=line, host=cur_entry, key=key, value=value))
            else:
                self.lines_.append(ConfigLine(line=line))

    def hosts(self):
        """
        Return the hosts found in the configuration.
        Returns
        -------
        Tuple of Host entries (including "*" if found)
        """
        return tuple(self.hosts_)

    def host(self, host):
        """
        Return the configuration of a specific host as a dictionary.
        Dictionary always contains lowercase versions of the attribute names.
        Parameters
        ----------
        host : the host to return values for.
        Returns
        -------
        dict of key value pairs, excluding "Host", empty map if host is not found.
        """
        if host in self.hosts_:
            vals = defaultdict(list)
            for k, value in [(x.key.lower(), x.value) for x in self.lines_
                             if x.host == host and x.key.lower() != "host"]:
                vals[k].append(value)
            flatten = lambda x: x[0] if len(x) == 1 else x
            return {k: flatten(v) for k, v in vals.items()}
        return {}

    def set(self, host, **kwargs):
        """
        Set configuration values for an existing host.
        Overwrites values for existing settings, or adds new settings.
        Parameters
        ----------
        host : the Host to modify.
        **kwargs : The new configuration parameters
        """
        self.__check_host_args(host, kwargs)

        def update_line(key, value):
            """Produce new config line"""
            return "  %s %s" % (key, value)

        for key, values in kwargs.items():
            if type(values) not in [list, tuple]:  # pylint: disable=unidiomatic-typecheck
                values = [values]

            lower_key = key.lower()
            update_idx = [idx for idx, x in enumerate(self.lines_)
                          if x.host == host and x.key.lower() == lower_key]
            extra_remove = []
            for idx in update_idx:
                if values:  # values available, update the line
                    value = values.pop()
                    self.lines_[idx].line = update_line(self.lines_[idx].key, value)
                    self.lines_[idx].value = value
                else:                # no more values available, remove the line
                    extra_remove.append(idx)

            for idx in reversed(sorted(extra_remove)):
                del self.lines_[idx]

            if values:
                mapped_key = _remap_key(key)
                max_idx = max([idx for idx, line in enumerate(self.lines_) if line.host == host])
                for value in values:
                    self.lines_.insert(max_idx + 1, ConfigLine(line=update_line(mapped_key, value),
                                                               host=host, key=mapped_key,
                                                               value=value))

    def unset(self, host, *args):
        """
        Removes settings for a host.
        Parameters
        ----------
        host : the host to remove settings from.
        *args : list of settings to removes.
        """
        self.__check_host_args(host, args)
        remove_idx = [idx for idx, x in enumerate(self.lines_)
                      if x.host == host and x.key.lower() in args]
        for idx in reversed(sorted(remove_idx)):
            del self.lines_[idx]

    def __check_host_args(self, host, keys):
        """Checks parameters"""
        if host not in self.hosts_:
            raise ValueError("Host %s: not found" % host)

        if "host" in [x.lower() for x in keys]:
            raise ValueError("Cannot modify Host value")

    def rename(self, old_host, new_host):
        """
        Renames a host configuration.
        Parameters
        ----------
        old_host : the host to rename.
        new_host : the new host value
        """
        if new_host in self.hosts_:
            raise ValueError("Host %s: already exists." % new_host)
        for line in self.lines_:  # update lines
            if line.host == old_host:
                line.host = new_host
                if line.key.lower() == "host":
                    line.value = new_host
                    line.line = "Host %s" % new_host
        self.hosts_.remove(old_host)  # update host cache
        self.hosts_.add(new_host)

    def add(self, host, **kwargs):
        """
        Add another host to the SSH configuration.
        Parameters
        ----------
        host: The Host entry to add.
        **kwargs: The parameters for the host (without "Host" parameter itself)
        """
        if host in self.hosts_:
            raise ValueError("Host %s: exists (use update)." % host)
        self.hosts_.add(host)
        self.lines_.append(ConfigLine(line="", host=None))
        self.lines_.append(ConfigLine(line="Host %s" % host, host=host, key="Host", value=host))
        for k, v in kwargs.items():
            if type(v) not in [list, tuple]:
                v = [v]
            mapped_k = _remap_key(k)
            for value in v:
                self.lines_.append(ConfigLine(line="  %s %s" % (mapped_k, str(value)), host=host, key=mapped_k, value=value))
        self.lines_.append(ConfigLine(line="", host=None))

    def remove(self, host):
        """
        Removes a host from the SSH configuration.
        Parameters
        ----------
        host : The host to remove
        """
        if host not in self.hosts_:
            raise ValueError("Host %s: not found." % host)
        self.hosts_.remove(host)
        # remove lines, including comments inside the host lines
        host_lines = [ idx for idx, x in enumerate(self.lines_) if x.host == host ]
        remove_range = reversed(range(min(host_lines), max(host_lines) + 1))
        for idx in remove_range:
            del self.lines_[idx]

    def config(self):
        """
        Return the configuration as a string.
        """
        return "\n".join([x.line for x in self.lines_])

    def write(self, path):
        """
        Writes ssh config file
        Parameters
        ----------
        path : The file to write to
        """
        with open(path, "w") as fh_:
            fh_.write(self.config())



from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
from .core.exc import ToolkitError
from .controllers.base import Base

# configuration defaults
CONFIG = init_defaults('dt')
CONFIG['dt']['foo'] = 'bar'


class Toolkit(App):
    """Data Toolkit primary application."""

    class Meta:
        label = 'dt'

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
            'jinja2',
        ]

        # configuration handler
        config_handler = 'yaml'

        # configuration file suffix
        config_file_suffix = '.yml'

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'jinja2'

        # register handlers
        handlers = [
            Base
        ]


class ToolkitTest(TestApp,Toolkit):
    """A sub-class of Toolkit that is better suited for testing."""

    class Meta:
        label = 'dt'


def main():
    with Toolkit() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except ToolkitError as e:
            print('ToolkitError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# TODO: early remote code execution.

import os
import sys
import sh
import fabric
import uuid


def main():
    if os.getcwd() == os.path.expanduser('~'):
        print('NO')
        exit(1)
    addr = sys.argv[1]
    del sys.argv[1]
    path = '/tmp/{}'.format(uuid.uuid4())
    env = []
    while '=' in sys.argv[1]:
        env.append(sys.argv[1].split('='))
        del sys.argv[1]
    env = dict(env)

    print(sh.rsync('-a', '-v', '-z', '--exclude', '__pycache__', '--exclude', '*.swp', '--exclude', '.git', '{}/'.format(os.getcwd()), '{}:{}'.format(addr, path)))

    cmd = 'cd {} && {}'.format(path, ' '.join(sys.argv[1:]))

    connection = fabric.Connection(addr, inline_ssh_env=True)
    connection.client.load_system_host_keys()
    connection.run(cmd, env=env)

    print(sh.rsync('-a', '-v', '-z', '{}:{}/logs/'.format(addr, path), os.getcwd() + '/logs'))


if __name__ == '__main__':
    main()


import os
import subprocess as sp
import pandas as pd
import json 

term_wdith = os.get_terminal_size().columns

pd.set_option('display.max_colwidth',term_wdith - 20)

def gcp_ls():
    # TODO: non-preemptible instances have wrong formatting.
    instances_list = sp.check_output('gcloud compute instances list'.split(' '))
    stringified_instances = str(instances_list).split('\\n')
    df = pd.DataFrame([ d.split()  for d in stringified_instances[:-1] ])
    # df.rename(columns=df.iloc[0]).drop(df.index[0])
    df = df.T.set_index(0).T
    return df


def scp(inst_id: int, rpath: str):
    gcp_instances = gcp_ls()
    tar_ip = gcp_instances.iloc[inst_id-1].EXTERNAL_IP
    tar_name = gcp_instances.iloc[inst_id-1].iloc[0]
    sftp = os.path.abspath(__file__).split('/')[:-2] + ['templates', 'sftp.json']

    scp_json = json.load(open('/'.join(sftp)))
    scp_json['host'] = tar_ip
    scp_json['name'] = tar_name
    scp_json['remotePath'] = rpath
    
    if not os.path.exists('.vscode'): os.mkdir('.vscode')
    tar_file = open(f"{os.getcwd()}/.vscode/sftp.json",'w')

    json.dump(scp_json, tar_file)
    print(f'Succesfully written {tar_name} (IP: {tar_ip}) into {str(tar_file)}')
    # import ipdb; ipdb.set_trace()


from ..ext.hist import hist_tail
import os

def run_command(n_lines: list):
    # TODO: propagate that last line in n_lines is the # of lines used
    # maybe add a check for 1 num to use default (20)
    df = hist_tail(n_lines[-1])
    exec_line = ''
    for cmd in n_lines[:-1]:
        exec_line += f"{df.iloc[cmd].Command}"
        exec_line += ' && '

    print(f"Running {exec_line[:-4]}")
    os.system(exec_line[:-4])
    # print(exec_line[:-4])


import sys
import pandas as pd
import datetime as dt
import subprocess as sp
from Levenshtein import distance as dist
import os
import humanize

term_wdith = os.get_terminal_size().columns

pd.set_option('display.max_colwidth',term_wdith - 20)

# TODO: treat 'cd' in a special way s.t. we can generate the absolute path wherever possible
# TODO: treat 'cd' in a specical way so that it can keep track of location.
FILTER_TERMS = ['ls','install','dt hist', 'git ', 'dt ', 'code',
                'ipython','htop','git push','git pull','git stash',
                'tmux', 'nvidia-smi', 'ping']

def dt_conv(l: str):
    try:
        # TODO: correct timezone? why does midnight show up as 4 pm?
        pd_dt_repr = pd.to_datetime(dt.datetime.fromtimestamp(int(l[1].strip())))
        # ret_val = '_'.join(str(pd_dt_repr).split(' ')[::-1])
        return  humanize.naturaldelta(pd_dt_repr)
    except ValueError as e:
        # TODO: why does this occur
        # print(e)
        pass


def filter_term(term: tuple):
    # Logic: all filter terms need to _not_ match
    if 'cd' in term[0][:2]:
        try:
            if term[0].split(' ')[1][0] not in ['~','/']:
                return False
            else:
                return True
        except:
            print(f"Error with {term[0]}")

    return all([ft not in term[0] for ft in FILTER_TERMS])

def hist_tail(n_lines=20):
    shell = os.environ['SHELL'].split('/')[-1]

    if shell=='sh': shell = 'bashrc'

    # checks that we can find 
    zsh_hist = f"{os.path.expanduser('~')}/.{shell}_history"
    if not os.path.exists(zsh_hist):
        # dumb hack for bashrc
        # zsh_hist = f"{os.path.expanduser('~')}/.{shell}rc"
        # if not.path.exists(zsh_hist): raise NotImplementedError('Not sure where to look')
    # else:
        raise NotImplementedError('Not sure where to look')
    split_f = str(sp.check_output(['tail',zsh_hist,'-n',str(n_lines * 5)])).split(':')

    # TODO: \n -> \\n for Mac, check if this is an issue on Linux
    # Work Ubuntu 20.04 so far so good
    parsed_commands = [ ''.join(l.replace('\\n','').strip().split(';')[-1:]) for l in split_f if '\\n' in l]
    parsed_dt = [ t for t in split_f if '\\n' not in t]

    parsed = list(zip(parsed_commands, parsed_dt))

    # remove FILTER_TERMS 
    cmds = [ t for t in parsed if filter_term(t) ]

    # identify Levenstein distance
    lev_dist = [ dist(cmds[i][0],cmds[i+1][0]) for i in range(len(cmds)-1) ]

    # TODO: make more efficent
    # Lev dist with previous terms (1 approx to catch typos)
    ld = lev_dist + [9999]
    # dist_dict = { k:v for k,v in zip(cmds,ld) }
    dist_dict = { k:v for k,v in zip(cmds,ld) if v>4 }
    # TODO check if ordered before
    last_list = list(dist_dict.keys())[-n_lines:]

    # constructs DF with one line per command
    last_cmds = [ l[0].replace('\\','') for l in last_list ]
    # TODO: the time seems to be wrong still
    last_dt = [ dt_conv(l) for l in last_list]
    df = pd.DataFrame([last_cmds,last_dt]).T
    df.columns = ['Command', 'Time']

    return df

def hist_show(line: str):
    # TODO: some weird number shows up? Why?
    history = hist_tail()
    return history.loc[8].Command


import torch
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from typing import Optional, Callable
import cv2

# np.set_printoptions(precision=3)

# Return Jaccard index, or Intersection over Union (IoU) value
def IoU(preds, targs, eps:float=1e-8):
    """Computes the Jaccard loss, a.k.a the IoU loss.
    Notes: [Batch size,Num classes,Height,Width]
    Args:
        targs: a tensor of shape [B, H, W] or [B, 1, H, W].
        preds: a tensor of shape [B, C, H, W]. Corresponds to
            the raw output or logits of the model. (prediction)
        eps: added to the denominator for numerical stability.
    Returns:
        iou: the average class intersection over union value
             for multi-class image segmentation
    """
    num_classes = preds.shape[1]

    # Single class segmentation?
    if num_classes == 1:
        true_1_hot = torch.eye(num_classes + 1)[targs.squeeze(1)]
        true_1_hot = true_1_hot.permute(0, 3, 1, 2).float()
        true_1_hot_f = true_1_hot[:, 0:1, :, :]
        true_1_hot_s = true_1_hot[:, 1:2, :, :]
        true_1_hot = torch.cat([true_1_hot_s, true_1_hot_f], dim=1)
        pos_prob = torch.sigmoid(preds)
        neg_prob = 1 - pos_prob
        probas = torch.cat([pos_prob, neg_prob], dim=1)

    # Multi-class segmentation
    else:
        # Convert target to one-hot encoding
        # true_1_hot = torch.eye(num_classes)[torch.squeeze(targs,1)]
        true_1_hot = torch.eye(num_classes)[targs.squeeze(1)]

        # Permute [B,H,W,C] to [B,C,H,W]
        true_1_hot = true_1_hot.permute(0, 3, 1, 2).float()

        # Take softmax along class dimension; all class probs add to 1 (per pixel)
        probas = F.softmax(preds, dim=1)

    true_1_hot = true_1_hot.type(preds.type())

    # Sum probabilities by class and across batch images
    dims = (0,) + tuple(range(2, targs.ndimension()))
    intersection = torch.sum(probas * true_1_hot, dims) # [class0,class1,class2,...]
    cardinality = torch.sum(probas + true_1_hot, dims)  # [class0,class1,class2,...]
    union = cardinality - intersection
    iou = (intersection / (union + eps)).mean()   # find mean of class IoU values
    return iou

def mean_iou(
        input: torch.Tensor,
        target: torch.Tensor,
        num_classes: int,
        eps: Optional[float] = 1e-6) -> torch.Tensor:
    r"""Calculate mean Intersection-Over-Union (mIOU).

    The function internally computes the confusion matrix.

    Args:
        input (torch.Tensor) : tensor with estimated targets returned by a
          classifier. The shape can be :math:`(B, *)` and must contain integer
          values between 0 and K-1.
        target (torch.Tensor) : tensor with ground truth (correct) target
          values. The shape can be :math:`(B, *)` and must contain integer
          values between 0 and K-1, whete targets are assumed to be provided as
          one-hot vectors.
        num_classes (int): total possible number of classes in target.

    Returns:
        torch.Tensor: a tensor representing the mean intersection-over union
        with shape :math:`(B, K)` where K is the number of classes.
    """
    from kornia.utils.metrics import *
    np.set_printoptions(precision=3)

    input = torch.argmax(input, axis=1 )
    if not torch.is_tensor(input) and input.dtype is not torch.int64:
        raise TypeError("Input input type is not a torch.Tensor with "
                        "torch.int64 dtype. Got {}".format(type(input)))
    if not torch.is_tensor(target) and target.dtype is not torch.int64:
        raise TypeError("Input target type is not a torch.Tensor with "
                        "torch.int64 dtype. Got {}".format(type(target)))
    if not input.shape == target.shape:
        raise ValueError("Inputs input and target must have the same shape. "
                         "Got: {} and {}".format(input.shape, target.shape))
    if not input.device == target.device:
        raise ValueError("Inputs must be in the same device. "
                         "Got: {} - {}".format(input.device, target.device))
    if not isinstance(num_classes, int) or num_classes < 2:
        raise ValueError("The number of classes must be an intenger bigger "
                         "than two. Got: {}".format(num_classes))
    # we first compute the confusion matrix
    conf_mat: torch.Tensor = confusion_matrix(input, target, num_classes)

    # compute the actual intersection over union
    sum_over_row = torch.sum(conf_mat, dim=1)
    sum_over_col = torch.sum(conf_mat, dim=2)
    conf_mat_diag = torch.diagonal(conf_mat, dim1=-2, dim2=-1)
    denominator = sum_over_row + sum_over_col - conf_mat_diag

    # NOTE: we add epsilon so that samples that are neither in the
    # prediction or ground truth are taken into account.
    ious = (conf_mat_diag + eps) / (denominator + eps)

    return np.array(ious.cpu())

def center_crop(path):
    im = Image.open(str(path))

    max_val = min(im.shape)

    new_width, new_height = max_val, max_val
    width, height = im.size   # Get dimensions

    left = (width - new_width)/2
    top = (height - new_height)/2
    right = (width + new_width)/2
    bottom = (height + new_height)/2

    # Crop the center of the image
    im = im.crop((left, top, right, bottom))
    return np.array(im)

def plot_result_comparison(standard_path,
                            learn: torch.nn.Module,
                            label_fcn: Callable,
                            resize: tuple=None,
                            crop_center: bool=False):
    '''
    Plots comparison of results using `learner` that is defined globally.

        :stanndard_path: Path-object

    '''
    from kornia.utils.metrics import *
    # grab images
    display_img = np.array(Image.open(standard_path))
    mask = plt.imread(label_fcn(standard_path))

    # resize if necessary
    if crop_center:
        display_img = center_crop(standard_path)
        mask = center_crop(label_fcn(standard_path))
    if resize:
        display_img = cv2.resize(display_img, dsize=(resize[0], resize[1]), interpolation=cv2.INTER_LINEAR)
        mask = cv2.resize(mask, dsize=(resize[0], resize[1]), interpolation=cv2.INTER_LINEAR)

    # get results
    res = learn.predict(display_img)

    plt.subplots(131, figsize=(24,8))
    plt.subplot(131)
    plt.title('Original')
    plt.imshow(display_img)
    plt.subplot(132)
    plt.title('Target')
    plt.imshow(mask)
    plt.subplot(133)
    plt.title('Prediction')
    plt.imshow(res[0])
    plt.show()





from setuptools import setup, find_packages
from dt.core.version import get_version

VERSION = get_version()

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='data-toolkit',
    version=VERSION,
    description='ML & data helper code!',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Jakub Langr',
    author_email='james.langr@gmail.com',
    url='https://github.com/johndoe/myapp/',
    license='(c) Jakub Langr',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'dt': ['templates/*', 'dt/scripts/*']},
    data_files=[('',["dt/ext/zshrc.txt"])],
    include_package_data=True,
    python_requires='>=3.6',
    entry_points="""
        [console_scripts]
        dt = dt.main:main
    """,
    install_requires=[
        # 'sentry_sdk',
        'python-Levenshtein',
        'github3.py',
        'cement',
        'humanize',
        'jinja2',
        'pyyaml',
        'colorlog',
        'gputil',
    ]
)



from cement import Controller, ex
from cement.utils.version import get_version_banner
import ipdb
from ..core.version import get_version
from pprint import pprint as pp
import os

VERSION_BANNER = """
ML & data helper code! %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        # text displayed at the top of --help output
        description = 'ML & data helper code!'

        # text displayed at the bottom of --help output
        epilog = 'Usage: dt command [args] [kwargs]'

        # controller level arguments. ex: 'dt --version'
        arguments = [
            ### add a version banner
            ( [ '-v', '--version' ],
              { 'action'  : 'version',
                'version' : VERSION_BANNER } ),
        ]


    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()

    @ex(
        # TODO: implement `--home` for setting up home address
        help='AWS Security Groups helper functions',
        arguments=[
            (['action'],{
                "help" : "One of {update, show}",
                "action": "store"
            }),
            (['-n','--name'],{
                "help" : "Name of the SG to act on",
                "action": "store"
            }),
            (['-ip'],{
                "help" : "IP to add to security group",
                "action": "store"
            }),
            (['-p','--profile'],{
                "help" : "Which AWS profile should boto3 use",
                "action" : "store"
            }),
            (['-P','--ports'],{
                "help" : "Which ports should AWS SG set",
                "action" : "store"
            })
        ]
    )
    def sg(self):
        from ..ext.aws_sg import get_sg, update_sg_to_ip
        profile = self.app.pargs.__dict__.get('profile') or self.app.pargs.__dict__.get('p') or 'default'
        name = self.app.pargs.__dict__.get('name') or self.app.pargs.__dict__.get('n') or 'default'
        
        # which ports to change
        ports = self.app.pargs.__dict__.get('ports') or self.app.pargs.__dict__.get('P')
        if ports: ports = [ int(p) for p in ports.split(',') ]
        else: ports = [22, 8888, 8501, 8889, 5432]
        # if not isinstance(ports, list): ports = [ports]


        if self.app.pargs.action=='show':
            pp(get_sg(name, profile))

        elif self.app.pargs.action=='update':
            update_sg_to_ip(sg_name=self.app.pargs.name, profile=profile, ports=ports)

        elif self.app.pargs.action=='ls':
            from ..ext.aws_sg import list_security_groups
            df = list_security_groups()
            print(df)
            # for sg in df.GroupId.unique():
            #     print(f"GroupdId: {sg}")
            #     print(df[df.GroupId==sg])

        else:
            raise NotImplementedError('As action, chose one of {update, show}')

    @ex(
        help="GCP VM Instances helper commands.",
        arguments=[
            (['action'],{
                    "help" : "To show one of {ls, scp}.",
                    "action" : "store"
                }),
            (['--rpath'],{
                "help" : "What is the location on the remote server",
                "action": "store"
            }),
            (['-f','--filter'],{
                "help" : "One of {live, all}. Default: live",
                "action": "store"
            }),
            (['-id'],{
                "help": "InstanceId to be passed to ec2 start/stop",
                "action": "store"
            })
        ]
    )
    def gcp(self):
        from ..ext.gcp import gcp_ls, scp

        # profile = self.app.pargs.profile or 'default'
        # filt = self.app.pargs.filter or "live"

        if self.app.pargs.action=='ls':
            print(gcp_ls())

        # elif self.app.pargs.action=='update':
        #     print(update_ec2_ssh(profile))

        # elif self.app.pargs.action=='start':
        #     start_ec2(self.app.pargs.id, profile)

        elif self.app.pargs.action=='scp':
            scp(int(self.app.pargs.id), self.app.pargs.rpath)
            # stop_ec2(self.app.pargs.id, profile)

        else:
            raise NotImplementedError("Choose one of {ls, update, start, stop}")

    @ex(
        help="AWS EC2 helper commands.",
        arguments=[
            (['action'],{
                    "help" : "To show one of {ls, update}.",
                    "action" : "store"
                }),
            (['-p','--profile'],{
                "help" : "Which AWS profile should boto3 use.",
                "action": "store"
            }),
            (['-f','--filter'],{
                "help" : "One of {live, all}. Default: live",
                "action": "store"
            }),
            (['-id'],{
                "help": "InstanceId to be passed to ec2 start/stop",
                "action": "store"
            })
        ]
    )
    def ec2(self):
        from ..ext.aws_ec2 import get_ec2, start_ec2, stop_ec2, update_ec2_ssh

        profile = self.app.pargs.profile or 'default'
        filt = self.app.pargs.filter or "live"

        if self.app.pargs.action=='ls':
            print(get_ec2('live', profile, filt))

        elif self.app.pargs.action=='update':
            print(update_ec2_ssh(profile))

        elif self.app.pargs.action=='start':
            start_ec2(self.app.pargs.id, profile)

        elif self.app.pargs.action=='stop':
            stop_ec2(self.app.pargs.id, profile)

        else:
            raise NotImplementedError("Choose one of {ls, update, start, stop}")

    @ex(
        help="Display all available Fluidstack nodes.",
        arguments=[
            (['action'],{
                "help" : "To show one of {avail, ls, start, stop}.",
                "action" : "store"
            }),
            (['-stats'],{
            "help" : "One of {all, gpu} to show either all (df) or gpu (json)",
            "action": "store"
            }),
            (['-n_gpu'],{
                "help" : "# of gpus required to be on the machine",
                "action" : "store"
            }),
            (['-id'],{
                "help" : "ID of the container to work with",
                "action" : "store"
            })
        ]
    )
    def fls(self):
        from ..ext.fluidstack import avail_fl_nodes, show_running, \
            stop_container, start_container


        args = self.app.pargs
        if args.action=='avail':
            avail_fl_nodes(args.stats, args.n_gpu)

        elif args.action=='ls':
            show_running()

        elif args.action=='stop':
            stop_container(args.id)

        elif args.action=='start':
            start_container(args.id)

        else:
            raise NotImplementedError('Set the first parg to one of {avail, ls, start, stop}')


    @ex(
        help='Monitor lack of GPU activity. When there is none, runs -job',
        arguments=[
            (['-job'],{
                "help" : "Text of the command to run after no GPU activity",
                "action" : "store"
            }),
            (['-n'], {
                "help" : "Number of processes before 'no activity'.",
                "action" : "store"
            })
        ]
    )
    def monitor(self):
        from ..ext.monitor import submit
        submit(self.app.pargs.job)


    # number of files
    @ex(
        help='Displays the number of files',
        arguments=[
            (['-d'], {
            "help" : "which directory to show",
            "action" : "store" } )
        ]
    )
    def nf(self):
        loc = self.app.pargs.d or '.'
        os.system(f'ls {loc} | wc -l')

    @ex(
        help="Appends defaults: from zshrc.txt to ~/.bashrc or ~/.zshrc or tmux conf",
        arguments = [
            (['shell'], {
                "help" : "One of {sh, zsh, tmux}. Imports a DT version of zshrc into bashrc or zshrc. or tmux .conf into ~/",
                "action": "store"
            })
        ]
    )
    def load(self):
        # TODO: reload the file after.
        # TODO: load git aliases
        zshrc = os.path.abspath(__file__).split('/')[:-2] + ['ext', 'zshrc.txt']
        tmux = os.path.abspath(__file__).split('/')[:-2] + ['ext', 'tmux.conf']

        zshrc_path = os.path.join('/'.join(zshrc))
        tmux_path = os.path.join('/'.join(tmux))
        if self.app.pargs.shell == 'sh':
            os.system(f'cat {zshrc_path} >> ~/.bashrc ')
        elif self.app.pargs.shell == 'zsh':
            os.system(f'cat {zshrc_path} >> ~/.zshrc ')
        elif self.app.pargs.shell == 'tmux':
            os.system(f'cat {tmux_path} >> ~/.tmux.conf ')
        else:
            raise NotImplementedError("Try one of {sh, zsh, tmux}")


    @ex(
        help='Operations on CLI hist',
        arguments=[
            (['-s'],{
                "help" : "To show the full command if too long to fit in terminal",
                "action" : "store"
            }),
            (['-id'],{
                # TODO: likely delete
            "help" : "Which CLI snippet ID to run",
            "action": "store"
            }),
            (['-n'],{
                # TODO: probably better as mandatory
                "help" : "Which lines from the selected history to show/save",
                "action" : "store"
            })
        ]
    )
    def hist(self):
        if not self.app.pargs.s:
            from ..ext.hist import hist_tail
            import pprint as pp
            n_lines = int(self.app.pargs.n or 20)
            pp.pprint(hist_tail(n_lines))
        else: 
            from ..ext.hist import hist_show
            print(hist_show(self.app.pargs.n))

    @ex(
        help='Operations with snippets',
        arguments=[
            (['n'],{
                "help" : "Which snippet to run/create. Using commas to seperate ids"
                "e.g. `dt run 13,23`",
                "action" : "store"
            }),
            (['-ssh'],{
                "help" : "Whether the command (includes?)/is? an SSH command",
                "action": "store"
            }),
            (['-c'],{
                # TODO: probably better as mandatory
                "help" : "Which lines from the selected history to show/save",
                "action" : "store"
            })
            # TODO: what if non-default n lines has been pased to hist
        ]
    )
    def run(self):
        from ..ext.run import run_command
        n_lines = [int(x) for x in str(self.app.pargs.n).split(',')]
        run_command(n_lines)

    @ex(
        help='Operations on the SSH config',
        arguments=[
            (['action'],{
                "help" : "To show one of {set, ls, update}.",
                "action" : "store"
            }),
            (['-host'],{
            "help" : "Which SSH config host to change",
            "action": "store"
            }),
            (['-ip'],{
                "help" : "The IP to update in the ssh config as HostName",
                "action" : "store"
            })
        ]
    )
    def ssh(self):
        from ..ext.sshconf import read_ssh_config
        if self.app.pargs.action=='set':
            ssh_config_obj = read_ssh_config('/Users/jakub/.ssh/config')
            ssh_config_obj.set(self.app.pargs.host, HostName=self.app.pargs.ip)
            ssh_config_obj.write('/Users/jakub/.ssh/config')
            print(ssh_config_obj.config())

        elif self.app.pargs.action=='update':
            from ..ext.aws_ec2 import update_ec2_ssh
            profile ='default' # or  self.app.pargs
            print(update_ec2_ssh(profile))

        else:
            # raise NotImplementedError()
            print('Arg out of {set, update} provided assuming an SSH command.')
            ssh_config_obj = read_ssh_config('/Users/jakub/.ssh/config')
            target = self.app.pargs.action
            if not target in ssh_config_obj.hosts_: raise Exception(f'Host {target} not in configured hosts.')
            else:
                os.system('dt sg update')
                os.system(f'ssh {target}')

    @ex(
        help='Basic viz using streamlit for image comparisons',
        arguments = [
            (['images'], {
                'help' : 'Location of target images',
                'action' : "store"
            }),
            (['labels'], {
                'help' : 'Location of target labels',
                "action" : 'store'
            }),
            (['adapted'], {
                'help' : 'Location of domain adapted images',
                "action" : "store"
            })
            #     'action' : "store",
            #     "images" : '/efs/public_data/gta5/images',
            #     "labels" : '/efs/public_data/gta5/labels',
            #     "adapted" : '/efs/public_data/images'
            # })
        ]
    )
    def viz(self):
        args = self.app.pargs
        print(args.images)
        from ..ext.firelit import create_viz
        create_viz(args.images, args.labels, args.adapted)


    @ex(
        help='Operations with config.',
        arguments = [
            (['action'], {
                "help" : "One of {show, set, loc}: \
                *show*: show configs Sentry DNS etc. \
                *set*: Resets the config from CLI. \
                *loc*: shows the location of the config",
                'action' : "store",
            })
        ]
    )
    def config(self):
        # TODO: setup Data-toolkit specific config
        args = self.app.pargs

        # shows config
        if args.action=='show':
            for attr in dir(config):
                if not attr.startswith('_'):
                    print(f"{attr}: {getattr(config,attr)}")

        # shows the path of the config
        if args.action=='loc':
            print(os.path.abspath(__file__))

        # resets the config
        # TODO this config method needs programmatic acces for spot instances
        if args.action=='set':
            from ..ext.tracking import initialize
            initialize(overwrite=True)

        elif args.action not in ['set','loc', 'show']:
            raise NotImplementedError("Try one of {show, set, loc}")


    @ex(
        help='Tracks your ML code till finish',
        arguments=[
            (['flags'], {
                'help' : 'Which  script + flags you want to pass to the underlying python.'
                         'e.g. train.py --parallel --batch_size 256',
                'action' : "store"
            }),
            (['-ns'], {
                'help' : 'Shuts down the computer between in the shutdown_hours.'
                        'Set to 22-08. Default: True',
                "action" : 'store'
            }),
            (['-t'], {
                'help' : 'If we are testing / doing a dry run. Default: False',
                "action" : "store"
            }),
            (['-cfg'],{
                'help': "location of the config file",
                "action" : "store",
            }),
            (['-p'],{
                "help": "which Python interpreter to use."
                        " Default is determined by `which python3`",
                "action" : "store"
            })
        ]
    )
    def track(self):
        from ..ext.tracking import tracking, initialize, test_sentry
        ''' 'Tracks experiments by making a Sentry alert when a script finishes. '''
        args = self.app.pargs
        if self.app.pargs is not None:
            tracking(args.flags, args.t, args.ns, args.p)

    @ex(
        help='Test if sentry DSN is set up correctly.'
    )
    def test(self):
        from ..ext.tracking import initialize, test_sentry
        test_sentry()

    @ex(
        help='All Github related operations',
        arguments=[
             (['action'],{
                "help" : "To show one of {forks, contrib}.",
                "action" : "store"
            }),
            (['-url'],{
                # TODO: what if there's no forks to give?
                "help" : "Github URL to search for forks."
            })
        ]
    )
    def gh(self):
        if self.app.pargs.action=='forks':
            from ..ext.github import find_forks
            print(find_forks(self.app.pargs.url))
        if self.app.pargs.action=='contrib':
            from ..ext.github import find_contributors
            print(find_contributors(self.app.pargs.url))


    @ex(
        help='Manage TODOs using Google Keep',
        arguments=[
            (['action'],{
                "help" : "One of {add, ls, dn}",
                "action": "store"
            }),
            (['-t'],{
                "help": "Text of the TODO",
                "action": "store"
            }),
            (['-l'],{
                "help": "Length of time in hours this is expected to take",
                "action": "store"
            }),
            (['-p'],{
                "help" : "Project name. One of {DA, SL}",
                "action" : 'store'
            }),
            (['-i'],{
                "help" : "Id of the finished task",
                "action": "store"
            }),
            (['-f'],{
                "help" : "Filter TODOs for ls",
                "action": "store"
            })
        ]
    )
    def td(self):
        from ..ext.gkeep import add_todo, list_todos, done_todo
        project = self.app.pargs.p or 'SL'

        if self.app.pargs.action=='add':
            add_todo(self.app.pargs.t, self.app.pargs.l, self.app.pargs.p)
        elif self.app.pargs.action=='ls':
            list_todos(self.app.pargs.f or True,
                       self.app.pargs.p)
        elif self.app.pargs.action=='dn':
            done_todo(self.app.pargs.i)

        else:
            raise NotImplemented("Action must be one of {add, ls, dn}")

    @ex(
        help='Execute a Python command across all files in current dir.',
        arguments=[
            (['command'],{
                "help": "what python command to run",
                "action": "store"
            }),
            (['-f'],{
                "help" : "filter all files in in this directory and below by this expression.",
                "action" : "store"
            })
        ]
    )
    def py(self):
        from ..ext.python_exec import execute
        execute(self.app.pargs.command, self.app.pargs.f)